#-*- Mode: Perl -*-

## File: OSC/Socket/Send.pm
## Authors:
##   jdl <jdl@xdv.org> wrote the original sendOSC.pm
##   Bryan Jurish <moocow@ling.uni-potsdam.de> : added OO interface
## Description:
##  + OSC send interface for perl
##======================================================================

package OSC::Send;

use OSC qw(:all);

@ISA = qw(OSC);

our %DEFAULTS =
  (
   host => 'localhost',
   port => 5151,
   typetags => 1,
  );

##======================================================================
## Generic Constructor
##======================================================================

## $sock = OSC::Send->new(%args);
##   + args:
##     (
##      host     => $host,
##      port     => $port,
##      typetags => $bool,
##     )
sub new {
  my $that = shift;
  $host = 'localhost' if (!defined($host));
  $port = 5151 if (!$port);
  my $this = {
	      %DEFAULTS,
	      ref($that) ? %$that : qw(),
	      @_,
	     };
  return bless $this, ref($that)||$that;
}

## undef = $sock->send(@msg);
## undef = $sock->print(@msg);
##   + sends @msg (empty-string-joined) to the socket
*print = \&send;
sub send {
  my $this = shift;
  #OSC::sendOSC([$0, '-notypetags', '-h', @$this{qw(host port)}, join('',@_)]);
  OSC::sendOSC([$0,
		($this->{typetags} ? qw() : '-notypetags'),
		'-h', @$this{qw(host port)},
		join('',@_)]);
}


##======================================================================
## sendOSC remnants (still needed??)

###########################
# generate and return fake argument vector for call to sendOSC function
#  + v0.02: rewrote -moo
sub make_sendargv() {
  my ($host,$port) = @_;
  return ($0, '-h', $host, $port);
}

##-- what the heck do I need this for? -moo
sub OSCsrv_login {
  my ($h,$p,$n) = @_;
  @argv = &make_sendargv($h,$p);
  $osc_message = "/reg,5252,$n";
  print "login--> '$osc_message'\n";
  push @argv,$osc_message;
  #      foreach $bl (@argv) {
  #	print $bl,"\n";
  #      }
  sendOSC::HartCoreMode($#argv + 1,\@argv);
  @argv = ();
}

##-- what the heck do I need this for? -moo
sub OSCsrv_logout {
  my ($h,$p,$n) = @_;
  @argv = &make_sendargv($h,$p);
  $osc_message = "/unreg,5252,$n";
  print "logout--> '$osc_message'\n";
  push @argv,$osc_message;
  sendOSC::HartCoreMode($#argv + 1,\@argv);
  @argv = ();

}



1;
__END__

##======================================================================
## Docs
=pod

=head1 NAME

OSC::Send - Perl OO wrapper for OSC clients.

=head1 SYNOPSIS

 use OSC::Send;

 # Create a socket...
 $sock = OSC::Send->new(host=>'localhost',port=>5151);

 # ... and send some data
 $sock->send("/foo,a,b,c");

 # (try the above with 'dumpOSC 5151' running in another shell)

=cut

##======================================================================
## Description
=pod

=head1 DESCRIPTION

OSC::Send provides a Perl OO wrapper for OSC clients.

=cut

##======================================================================
## Footer
=pod

=head1 ACKNOWLEDGEMENTS

perl by Larry Wall.

=head1 KNOWN BUGS

None known.

=head1 AUTHOR(s)

jdl E<lt>jdl@xdv.orgE<gt> wrote the original OSC_perl.

Bryan Jurish E<lt>jurish@ling.uni-potsdam.deE<gt> wrote these wrappers.

=head1 SEE ALSO

perl(1),
sendOSC(1),
dumpOSC(1),
OSC(3pm),
OSC::Dump(3pm).

=cut
